#ifndef Lib_Array_h
#define Lib_Array_h

#include <vector>
#include <Lib/Debug.h>

namespace Lib {

// Array handles an hyperparallelepiped of cells of class T.

template <class T>
class Array {
public:
  typedef typename std::vector< std::pair<unsigned int,unsigned int> > Cut;
  typedef typename std::vector<unsigned int> UInts;
  typedef typename std::vector<T>::iterator VecIt;
  typedef typename std::vector<T>::const_iterator ConsVecIt;
public:
  Array() {
    Debug::increment("Lib::Array<T>");
  }
  Array(const UInts& aOrders) {
    Debug::increment("Lib::Array<T>");
    configure(aOrders);
  }
  Array(unsigned int aDimension,unsigned int aOrder) { 
    // A hypercube of dimension "aDimension" and size "aOrder".
    Debug::increment("Lib::Array<T>");
    UInts orders(aDimension);
    for(unsigned int index=0;index<aDimension;index++) orders[index] = aOrder;
    configure(orders);
  }
  Array(const Array& aFrom)
  :fOrders(aFrom.fOrders)
  ,fOffsets(aFrom.fOffsets)
  ,fVector(aFrom.fVector)
  ,fIs(aFrom.fIs){
    Debug::increment("Lib::Array<T>");
  }
  virtual ~Array() {
    Debug::decrement("Lib::Array<T>");
  }
  inline Array& operator=(const Array& aFrom) {
    fOrders = aFrom.fOrders;
    fOffsets = aFrom.fOffsets;
    fVector = aFrom.fVector;
    fIs = aFrom.fIs;
    return *this;
  }
  inline void copy(const Array& aFrom) {
    fOrders = aFrom.fOrders;
    fOffsets = aFrom.fOffsets;
    fVector = aFrom.fVector;
    fIs = aFrom.fIs;
  }
  inline void clear() {
    fOrders.clear();
    fOffsets.clear();
    fVector.clear();
    fIs.clear();
  }
  inline bool configure(const UInts& aOrders) {
    fOrders = aOrders;
    unsigned int dim = fOrders.size();
    if(dim==0) {
      clear();
      return false;
    }
    unsigned int size = 1;
    for(unsigned int index=0;index<dim;index++) {
      //if(fOrders[index]<0) {
      //clear();
      //return false;
      //}
      size *= fOrders[index];
    }
    fVector.resize(size,zero());
    fOffsets.resize(dim,0);
    fOffsets[0] = 1;
    for(unsigned int iaxis=1;iaxis<dim;iaxis++) 
      fOffsets[iaxis] = fOffsets[iaxis-1] * fOrders[iaxis-1];
    fIs.resize(dim,0);
    return true;
  }
  inline unsigned int dimension() const { return fOrders.size();}
  inline const UInts& orders() const { return fOrders;}
  inline unsigned int size() const {return fVector.size();}
  inline bool setValue(const UInts& aIs,const T& aValue) {
    unsigned int off = 0;
    if(!offset(aIs,off)) return false;
    fVector[off] = aValue;
    return true;
  }
  inline bool value(const UInts& aIs,T& aValue) const {
    unsigned int off = 0;
    if(!offset(aIs,off)) {
      aValue = 0;
      return false;
    }
    aValue = fVector[off];
    return true;
  }
  inline void reset() {
    for(VecIt it=fVector.begin();it!=fVector.end();++it) *it = 0;
  }
  inline const std::vector<T>& vector() const { return fVector;}
  inline std::vector<T>& vector() { return fVector;}
  inline bool fill(const std::vector<T>& aValues,Cut* aCut = 0) {
    unsigned int dsize = aValues.size();
    unsigned di = 0;
    unsigned int index = 0;
    for(VecIt it=fVector.begin();it!=fVector.end();++it,index++) {
      if(!aCut || (aCut && accept(index,*aCut)) ) {
        if(di>=dsize) return false; //aValues exhausted too early
        *it = aValues[di];
        di++;
      }
    }
    return true;
  }
  // Related to other Array :
  inline bool equal(const Array& aArray) const {
    if(fOrders!=aArray.fOrders) return false;
    ConsVecIt it = fVector.begin();
    ConsVecIt ait = aArray.fVector.begin();
    for(;it!=fVector.end();++it,++ait) {
      if((*it)!=(*ait)) return false;
    }
    return true;
  }
  inline bool equal(const Array& aArray,T aEpsilon) const {
    if(fOrders!=aArray.fOrders) return false;
    ConsVecIt it = fVector.begin();
    ConsVecIt ait = aArray.fVector.begin();
    for(;it!=fVector.end();++it,++ait) {
      T diff = (*it) - (*ait);      
      if(diff<0) diff *= -1;
      if(diff>=aEpsilon) return false;
    }
    return true;
  }
  inline bool isProportional(const Array& aArray,T& aFactor) const {
    // If true, then : aArray = aFactor * this.
    aFactor = zero();
    if(fOrders!=aArray.fOrders) return false;
    bool first = true;
    ConsVecIt it = fVector.begin();
    ConsVecIt ait = aArray.fVector.begin();
    for(;it!=fVector.end();++it,++ait) {
             if( ((*it)==zero()) && ((*ait)==zero())) {
        continue;
      } else if( ((*it)!=zero()) && ((*ait)==zero())) {
        return false;
      } else if( ((*it)==zero()) && ((*ait)!=zero())) {
        return false;
      } else {
        if(first) {
          aFactor = (*ait)/(*it);
          first = false;
        } else {
          if((*ait)!=(*it)*aFactor) return false;
        }
      }
    }
    return true;
  }
  inline bool add(const Array& aArray,Cut* aCut = 0) {
    if(fOrders!=aArray.fOrders) return false;
    VecIt it = fVector.begin();
    ConsVecIt ait = aArray.fVector.begin();
    unsigned int index = 0;
    for(;it!=fVector.end();++it,++ait,index++) {
      if(!aCut || (aCut && accept(index,*aCut)) ) {
        (*it) += (*ait);
      }
    }
    return true;
  }
  inline bool subtract(const Array& aArray) {
    if(fOrders!=aArray.fOrders) return false;
    VecIt it = fVector.begin();
    ConsVecIt ait = aArray.fVector.begin();
    for(;it!=fVector.end();++it,++ait) (*it) -= (*ait);
    return true;
  }
  inline bool multiply(const Array& aArray) {
    if(fOrders!=aArray.fOrders) return false;
    VecIt it = fVector.begin();
    ConsVecIt ait = aArray.fVector.begin();
    for(;it!=fVector.end();++it,++ait) (*it) *= (*ait);
    return true;
  }
  inline bool divide(const Array& aArray) {
    if(fOrders!=aArray.fOrders) return false;
    bool status = true;
    VecIt it = fVector.begin();
    ConsVecIt ait = aArray.fVector.begin();
    for(;it!=fVector.end();++it,++ait) {
      if((*ait)==zero())  {
        (*it) = zero(); //PAW convention.
        status = false;
      } else {
        (*it) /= (*ait);
      }
    }
    return status;
  }
  inline bool contract(const Array& aArray,T& aValue) const {
    aValue = zero();
    if(fOrders!=aArray.fOrders) return false;
    ConsVecIt it = fVector.begin();
    ConsVecIt ait = aArray.fVector.begin();
    for(;it!=fVector.end();++it,++ait) aValue += (*it) * (*ait);
    return true;
  }
  //Else:
  inline void add(const T& aT,Cut* aCut = 0) {
    unsigned int index = 0;
    for(VecIt it = fVector.begin();it!=fVector.end();++it,index++) {
      if(!aCut || (aCut && accept(index,*aCut)) ) {
        (*it) += aT;
      }
    }
  }
  inline void multiply(const T& aT) {
    for(VecIt it = fVector.begin();it!=fVector.end();++it) (*it) *= aT;
  }
  inline bool divide(const T& aT) {
    if(aT==zero()) return false;
    for(VecIt it = fVector.begin();it!=fVector.end();++it) (*it) /= aT;
    return true;
  }
  inline bool invert() {
    bool status = true;
    T v_one = one();
    for(VecIt it = fVector.begin();it!=fVector.end();++it) {
      if((*it)==zero())  {
        (*it) = zero(); //PAW convention.
        status = false;
      } else {
        T value = (*it); 
        (*it) = v_one/value;
      }
    }
    return status;
  }
  inline bool offset(const UInts& aIs,unsigned int& aOffset) const {
    unsigned int dim = fOrders.size();
    aOffset = 0;
    if(aIs.size()!=dim) return false;
    if(dim==0) return false;
    for(unsigned int iaxis=0;iaxis<dim;iaxis++) { 
      unsigned int i = aIs[iaxis];
      if(i>=fOrders[iaxis]) {
        aOffset = 0;
        return false;
      }
      aOffset += i * fOffsets[iaxis];
    }
    return true;
  }
  inline bool indices(unsigned int aOffset,UInts& aIs) const {
    if(aOffset>=fVector.size()) return false;
    unsigned int dim = fOrders.size();
    unsigned int off = aOffset;
    for(int iaxis=dim-1;iaxis>=0;iaxis--) { 
      aIs[iaxis] = off/fOffsets[iaxis];
      off -= aIs[iaxis] * fOffsets[iaxis];
    }
    return true;
  }
  inline bool accept(unsigned int aIndex,const Cut& aCut) const {
    unsigned int dim = fOrders.size();
    if(aCut.size()!=dim) return false;
    if(!indices(aIndex,const_cast<UInts&>(fIs))) return false;
    bool good = true;
    for(unsigned iaxis=0;iaxis<dim;iaxis++) { 
      if(fIs[iaxis]<aCut[iaxis].first) {good = false;break;}
      if(fIs[iaxis]>aCut[iaxis].second) {good = false;break;}
    }
    return good;
  }
  inline static T zero() {
    //return (T)0;    
    return T(0);    
  }
  inline static T one() {
    //return (T)1;    
    return T(1);    
  }
  inline static T minus_one() {
    //return (T)-1;    
    return T(-1);    
  }
  inline static T two() {
    //return (T)2;    
    return T(2);    
  }
protected:
  UInts fOrders;
  UInts fOffsets;
  std::vector<T> fVector;
  UInts fIs;
};

template <class T>
class ADD : public Array<T> {
public:
  ADD(unsigned aOrder):Array<T>(2,aOrder){
    // aOrder is the matrix dimension, which must not be confused
    // with the Array dimension which is 2).
  }
  ADD(const ADD& aFrom):Array<T>(aFrom){
  }
  virtual ~ADD() {}
  inline ADD& operator=(const ADD& aFrom) {
    Array<T>::operator=(aFrom);
    return *this;
  }

  inline bool setOrder(unsigned int aOrder) {
    std::vector<unsigned int> orders(2);
    for(unsigned int index=0;index<2;index++) orders[index] = aOrder;
    return Array<T>::configure(orders);
  }

  inline unsigned int order() const { return Array<T>::fOrders[0];}

  inline T value(unsigned int aR,unsigned int aC) const { 
    //WARNING : range of aR, aC not protected.
    unsigned int ord = order();
    return Array<T>::fVector[aR + aC * ord];
  }

  inline bool value(unsigned int aR,unsigned int aC,T& aValue) const { 
    unsigned int ord = order();
    if((aR>=ord)||(aC>=ord)) {
      aValue = Array<T>::zero();
      return false;
    }
    aValue = Array<T>::fVector[aR + aC * ord];
    return true;
  }

  inline bool setValue(unsigned int aR,unsigned int aC,const T& aValue) { 
    unsigned int ord = order();
    if((aR>=ord)||(aC>=ord)) return false;
    Array<T>::fVector[aR + aC * ord] = aValue;
    return true;
  }

  inline void setValueNoCheck(unsigned int aR,
                              unsigned int aC,
                              const T& aValue){ 
    Array<T>::fVector[aR + aC * order()] = aValue;
  }

  inline T trace() const {
    unsigned int ord = order();
    T value = Array<T>::zero();
    for(unsigned int c=0;c<ord;c++) value += Array<T>::fVector[c+c*ord];
    return value;
  }

  inline bool mxmul(const ADD<T>& aMatrix,ADD<T>& aResult) const {
    // aResult = this * aMatrix
    unsigned int ord = order();
    if(aMatrix.order()!=ord) return false;
    if(aResult.order()!=ord) return false;
    for(unsigned int r=0;r<ord;r++) {
      for(unsigned int c=0;c<ord;c++) {
        T value = Array<T>::zero();
        for(unsigned int i=0;i<ord;i++) {
          value += 
            Array<T>::fVector[r+i*ord] * 
            aMatrix.Array<T>::fVector[i+c*ord];
        }
        aResult.Array<T>::fVector[r+c*ord] = value;
      }
    }
    return true;
  }

  inline virtual T determinant() const {
    unsigned int ord = order();
    if(ord==0) return Array<T>::zero();
    else if(ord==1) return value(0,0);
    else if(ord==2) return (value(0,0) * value(1,1) - 
                            value(1,0) * value(0,1)); //Optimize

    unsigned int rord = ord-1;
    std::vector<unsigned int> cs(rord);
    std::vector<unsigned int> rs(rord);

    T det = Array<T>::zero();
   {for(unsigned int i=0;i<rord;i++) {cs[i] = i+1;}}
    unsigned int c = 0;
    //if(c>=1) cs[c-1] = c-1;

   {for(unsigned int i=0;i<rord;i++) {rs[i] = i+1;}}
    bool sg = true; //c=0+r=0
    for(unsigned int r=0;r<ord;r++) {
      if(r>=1) rs[r-1] = r-1;
      T subdet = subDeterminant(rs,cs);
      if(sg) 
        det += value(r,c) * subdet;
      else
        det -= value(r,c) * subdet;
      sg = sg?false:true;
    }

    return det;
  }
  inline T subDeterminant(const std::vector<unsigned int>& aRs,
                          const std::vector<unsigned int>& aCs) const {
    //WARNING : to optimize, we do not check the content of aRs, aCs.
    unsigned int ord = aRs.size();
    if(ord==0) return Array<T>::zero();
    else if(ord==1) return value(aRs[0],aCs[0]);
    else if(ord==2) return (value(aRs[0],aCs[0]) * value(aRs[1],aCs[1]) -
                            value(aRs[1],aCs[0]) * value(aRs[0],aCs[1])); //Opt

    unsigned int rord = ord-1;
    std::vector<unsigned int> cs(rord);
    std::vector<unsigned int> rs(rord);

    T det = Array<T>::zero();
   {for(unsigned int i=0;i<rord;i++) {cs[i] = aCs[i+1];}}
    unsigned int c = 0;
    //if(c>=1) cs[c-1] = c-1;

   {for(unsigned int i=0;i<rord;i++) {rs[i] = aRs[i+1];}}
    bool sg = true; //c=0+r=0
    for(unsigned int r=0;r<ord;r++) {
      if(r>=1) rs[r-1] = aRs[r-1];
      T subdet = subDeterminant(rs,cs);
      if(sg)
        det += value(aRs[r],aCs[c]) * subdet;
      else
        det -= value(aRs[r],aCs[c]) * subdet;
      sg = sg?false:true;
    }

    return det;
  }

  inline bool invert(ADD<T>& aResult) const {
    //Generic invertion method.
    unsigned int ord = order();
    if(ord==0) return true;

    aResult.setOrder(ord);
    if(ord==1) {
      T v;
      if(!value(0,0,v)) return false;
      if(v==Array<T>::zero()) return false;
      if(!aResult.setValue(0,0,1./v)) return false;
      return true;
    }
  
    unsigned int rord = ord-1;
    std::vector<unsigned int> cs(rord);
    std::vector<unsigned int> rs(rord);
  
    // Get det with r = 0;
    T det = Array<T>::zero();
   {
   {for(unsigned int i=0;i<rord;i++) {rs[i] = i+1;}}
    unsigned int r = 0;
    //if(r>=1) rs[r-1] = r-1;
  
   {for(unsigned int i=0;i<rord;i++) {cs[i] = i+1;}}
    bool sg = true; //r=0+c=0
    for(unsigned int c=0;c<ord;c++) {
      if(c>=1) cs[c-1] = c-1;
      T subdet = subDeterminant(rs,cs);
      T sgn = sg ? Array<T>::one() : Array<T>::minus_one();
      det += value(r,c) * subdet * sgn;
      T value = subdet * sgn;
      aResult.setValueNoCheck(c,r,value);
      sg = sg?false:true;
    }}
  
    if(det==Array<T>::zero()) return false;
  
   {for(unsigned int c=0;c<ord;c++) {
      aResult.setValue(c,0,aResult.value(c,0)/det);
    }}
  
   {for(unsigned int i=0;i<rord;i++) {rs[i] = i+1;}}
    bool sgr = false; //r=1+c=0
    for(unsigned int r=1;r<ord;r++) {
      if(r>=1) rs[r-1] = r-1;
     {for(unsigned int i=0;i<rord;i++) {cs[i] = i+1;}}
      bool sg = sgr;
      for(unsigned int c=0;c<ord;c++) {
        if(c>=1) cs[c-1] = c-1;
        T subdet = subDeterminant(rs,cs);
        T sgn = sg ? Array<T>::one() : Array<T>::minus_one();
        T value = (subdet * sgn)/det;
        aResult.setValueNoCheck(c,r,value);
        sg = sg?false:true;
      }
      sgr = sgr?false:true;
    }
  
    return true;
  }

  inline void transpose() {
    unsigned int ord = order();
    for(unsigned int r=0;r<ord;r++) {
      for(unsigned int c=(r+1);c<ord;c++) {
        T vrc = value(r,c);
        T vcr = value(c,r);
        setValueNoCheck(r,c,vcr);
        setValueNoCheck(c,r,vrc);
      }
    }
  }

  inline bool isSymmetric() const {
    unsigned int ord = order();
    for(unsigned int r=0;r<ord;r++) {
      for(unsigned int c=(r+1);c<ord;c++) {
        if(value(r,c)!=value(c,r)) return false;
      }
    }
    return true;
  }
  inline bool isAntiSymmetric() const {
    unsigned int ord = order();
   {for(unsigned int r=0;r<ord;r++) {
      if(value(r,r)!=Array<T>::zero()) return false;
    }}
    for(unsigned int r=0;r<ord;r++) {
      for(unsigned int c=(r+1);c<ord;c++) {
        if(value(r,c)!=-value(c,r)) return false;
      }
    }
    return true;
  }

  inline void setIdentity() {
    unsigned int ord = order();
    Array<T>::reset();
    for(unsigned int c=0;c<ord;c++) setValueNoCheck(c,c,Array<T>::one());
  }

  //Factories :
  inline ADD symmetricPart() const {
    ADD res(*this);
    res.transpose();
    res.add(*this);
    res.divide(Array<T>::two());
    return res;
  }

  inline ADD antiSymmetricPart() const {
    ADD res(*this);
    res.transpose();
    res.multiply(Array<T>::minus_one());
    res.add(*this);
    res.divide(Array<T>::two());
    return res;
  }

  inline ADD& operator*=(const Lib::ADD<T>& aFrom) {
    Lib::ADD<T> res(order());
    /*bool status =*/ this->mxmul(aFrom,res); //FIXME : throw
    *this = res;
    return *this;
  }
  inline ADD& operator+=(const Lib::ADD<T>& aFrom) {
    /*bool status =*/ Array<T>::add(aFrom); //FIXME : throw
    return *this;
  }
  inline ADD& operator-=(const Lib::ADD<T>& aFrom) {
    /*bool status =*/ Array<T>::subtract(aFrom); //FIXME : throw
    return *this;
  }
};

}

#include <Lib/Out.h>
// Helpers Array<T> :
namespace Lib {

//////////////////////////////////////////////////////////////////////////////
template <class T>
bool contract(
 const Array<T>& aA
,unsigned int aIA
,const Array<T>& aB
,unsigned int aIB
,Array<T>& aR
)
//////////////////////////////////////////////////////////////////////////////
// Contract a tensor (aA) with a vector (aB) on indice aIA of aA and aIB of aB.
// aA.dimension must be > 1.
// aB.dimension must be > 1.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(&aR==&aA) return false;
  if(&aR==&aB) return false;

  if(aA.dimension()==0) return false;
  if(aB.dimension()==0) return false;
  if(aIA>=aA.dimension()) return false;
  if(aIB>=aB.dimension()) return false;
  if(aA.orders()[aIA]!=aB.orders()[aIB]) return false;

  unsigned int rdima = aA.dimension()-1;
  unsigned int rdimb = aB.dimension()-1;

  if((rdima+rdimb)==0) return false;

  std::vector<unsigned int> rorders(rdima+rdimb);
  unsigned int index;
  for(index=0;index<aIA;index++) 
    rorders[index] = aA.orders()[index];
  for(index=aIA;index<rdima;index++) 
    rorders[index] = aA.orders()[index+1];

  for(index=0;index<aIB;index++) 
    rorders[rdima+index] = aB.orders()[index];
  for(index=aIB;index<rdimb;index++) 
    rorders[rdima+index] = aB.orders()[index+1];

  if(!aR.configure(rorders)) return false;

  std::vector<unsigned int> ais(aA.dimension());
  std::vector<unsigned int> bis(aB.dimension());
  std::vector<unsigned int> ris(aR.dimension());

  //FIXME : optimize the below.
  unsigned int order = aA.orders()[aIA];
  unsigned int rsize = aR.size();

  std::vector<T>& rvec = aR.vector();

  for(unsigned int roffset=0;roffset<rsize;roffset++) {
    if(!aR.indices(roffset,ris)) return false;

    for(index=0;index<aIA;index++) ais[index] = ris[index];
    for(index=aIA;index<rdima;index++) ais[index+1] = ris[index];

    for(index=0;index<aIB;index++) bis[index] = ris[rdima+index];
    for(index=aIB;index<rdimb;index++) bis[index+1] = ris[rdima+index];

    T value = 0;
    for(index=0;index<order;index++) {
      ais[aIA] = index;
      T av;
      if(!aA.value(ais,av)) return false;

      bis[aIB] = index;
      T bv;
      if(!aB.value(bis,bv)) return false;

      value += av * bv;
    }

    rvec[roffset] = value;
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
template <class T>
bool swap(
 const Array<T>& aV
,unsigned int aI1
,unsigned int aI2
,Array<T>& aR
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(&aR==&aV) return false;

  unsigned int dim = aV.dimension();
  if(aI1>=dim) return false;
  if(aI2>=dim) return false;

  if(aI1==aI2) {
    aR.copy(aV);
    return true;
  }

  if(!aR.configure(aV.orders())) return false;

  std::vector<unsigned int> vis(aV.dimension());
  std::vector<unsigned int> ris(aV.dimension());

  const std::vector<T>& vvec = aV.vector();

  unsigned int size = aV.size();
  for(unsigned int offset=0;offset<size;offset++) {
    T value = vvec[offset];

    if(!aV.indices(offset,vis)) return false;
    unsigned int i = vis[aI1]; 
    vis[aI1] = vis[aI2]; 
    vis[aI2] = i; 
    if(!aR.setValue(vis,value)) return false;
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
template <class T> 
void print(
 Slash::Core::IWriter& aPrinter
,const Lib::Array<T>& aArray
,const std::string& aTitle
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Out out(aPrinter);
  if(aTitle!="") out << aTitle << Lib::endl;
  const std::vector<T>& vec = aArray.vector();
  typedef typename std::vector<T>::const_iterator ConsVecIt;
  for(ConsVecIt it = vec.begin();it!=vec.end();++it) {
    out << (*it) << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
template <class T> 
void diff(
 Slash::Core::IWriter& aPrinter
,const Array<T>& aA
,const Array<T>& aB
,T aEpsilon
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Out out(aPrinter);
  if(aA.orders()!=aB.orders()) {
    out << "Lib::Arrays::diff : not same orders" << Lib::endl;
    return;
  }
  bool header_done = false;
  unsigned int dim = aA.dimension();
  std::vector<unsigned int> is(dim);
  unsigned int vsize = aA.vector().size();
  for(unsigned int index=0;index<vsize;index++) {
    T diff = aA.vector()[index]-aB.vector()[index];      
    if(diff<0) diff *= -1;
    if(diff>=aEpsilon) {
      aA.indices(index,is);
      if(!header_done) {
        out << "Lib::Arrays::diff :" << Lib::endl;
        header_done = true;
      }
      for(unsigned int i=0;i<dim;i++) out << is[i] << " ";
      out << aA.vector()[index] << " " << aB.vector()[index] << Lib::endl;
    }
  }
}

} //Helpers end

// Helpers ADD<T> :
#include <Lib/sout.h>

template <class T>
Lib::ADD<T> operator-(const Lib::ADD<T>& a1,const Lib::ADD<T>& a2) {
  Lib::ADD<T> res(a1);
  res.subtract(a2);
  return res;
}
template <class T>
Lib::ADD<T> operator+(const Lib::ADD<T>& a1,const Lib::ADD<T>& a2) {
  Lib::ADD<T> res(a1);
  res.add(a2);
  return res;
}
template <class T>
Lib::ADD<T> operator*(const Lib::ADD<T>& a1,const Lib::ADD<T>& a2) {
  Lib::ADD<T> res(a1.order());
  a1.mxmul(a2,res);
  return res;
}

namespace Lib {

template <class T>
ADD<T> commutator(const ADD<T>& a1,const ADD<T>& a2) {
  /*
  ADD<T> res(a1.order());
  a1.mxmul(a2,res);
  ADD<T> tmp(a1.order());
  a2.mxmul(a1,tmp);
  res.subtract(tmp);
  return res;
  */
  return a1*a2-a2*a1;
}

template <class T>
ADD<T> antiCommutator(const ADD<T>& a1,const ADD<T>& a2) {
  /*
  ADD<T> res(a1.order());
  a1.mxmul(a2,res);
  ADD<T> tmp(a1.order());
  a2.mxmul(a1,tmp);
  res.add(tmp);
  return res;
  */
  return a1*a2+a2*a1;
}

template <class T>
void print(const ADD<T>& aMatrix,Slash::Core::IWriter& aPrinter,const std::string& aCMT) {
  Lib::Out out(aPrinter);
  if(aCMT!="") out << aCMT << Lib::endl;
  unsigned int ord = aMatrix.order();
  for(unsigned int r=0;r<ord;r++) {
    for(unsigned int c=0;c<ord;c++) {
      out << " " << Lib::sout(aMatrix.value(r,c));
    }
    out << Lib::endl;
  }
}

template <class T>
void checkInvert(
 const ADD<T>& aMatrix
,Slash::Core::IWriter& aPrinter
) {
  unsigned int ord = aMatrix.order();
  Lib::ADD<T> id(ord);
  id.setIdentity();
  Lib::ADD<T> inv(ord);
  if(!aMatrix.invert(inv)) return;
  //Lib::ADD<T> tmp(ord);
  //aMatrix.mxmul(inv,tmp);
  Lib::ADD<T> tmp = aMatrix * inv;
  if(!tmp.equal(id)) {
    Lib::Out out(aPrinter);
    Lib::print(aMatrix,aPrinter,"problem with inv of :");
    Lib::print(inv,aPrinter,"found :");
  }
}

template <class T>
ADD<T> exp(const ADD<T>& aMatrix,unsigned int aNumber = 100) {
  // result = I + M + M**2/2! + M**3/3! + ....
  ADD<T> result(aMatrix.order());
  result.setIdentity();
  ADD<T> m(aMatrix.order());     
  m.setIdentity();
  for(unsigned int i=1;i<aNumber;i++) {
    m = m * aMatrix; 
    m.multiply(1./double(i)); 
    result.add(m);
  }
  return result;
}

template <class T>
ADD<T> log(const ADD<T>& aMatrix,unsigned int aNumber = 100) {
  // result = (M-I) - (M-I)**2/2 + (M-I)**3/3 +...
  // Warning : cannot converge...
  unsigned int order = aMatrix.order();
  ADD<T> result(order);
  ADD<T> id(order);     
  id.setIdentity();
  ADD<T> x(aMatrix);     
  x.subtract(id);
  ADD<T> m(id);
  ADD<T> tmp(order);
  double fact = -1;
  for(unsigned int i=0;i<aNumber;i++) {
    m = m * x; 
    fact *= -1;
    tmp = m;
    tmp.multiply(fact/double(i+1)); 
    result.add(tmp);
  }
  return result;
}

template <class T>
ADD<double> abs(const ADD<T>& aMatrix,double(*aFunc)(T)) {
  const std::vector<T>& avec = aMatrix.vector();
  ADD<double> result(aMatrix.order());
  std::vector<double>& rvec = result.vector();
  unsigned int number = avec.size();
  for(unsigned int index=0;index<number;index++) 
    rvec[index] = aFunc(avec[index]);
  return result;
}

template <class T>
bool isLogConvergent(const ADD<T>& aMatrix,double(*aFunc)(T)) {
  unsigned int order = aMatrix.order();
  ADD<T> id(order);     
  id.setIdentity();
  ADD<T> s(aMatrix);
  s.subtract(id);
  ADD<double> a = abs(s,aFunc);
  double radius = 1./order;

  std::vector<double>& vec = a.vector();
  typedef typename std::vector<T>::iterator VecIt;
  for(VecIt it = vec.begin();it!=vec.end();++it) {
    if((*it)>=radius) {
      //::printf("debug : not convergent at : %d : %g rad %g\n",
        //index,a.value(index),radius);
      return false;
    }
  }

  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// To work with complex. ////////////////////////////////////////////////////
/// The below uses conj. /////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
template <class T> 
void conj(
 Lib::Array<T>& aArray
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<T>& vec = aArray.vector();
  typedef typename std::vector<T>::iterator VecIt;
  for(VecIt it = vec.begin();it!=vec.end();++it) {
    (*it) = conj(*it);     
  }
}
//////////////////////////////////////////////////////////////////////////////
template <class T> 
void dagger(
 Lib::ADD<T>& aArray
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  conj(aArray);
  aArray.transpose();
}

}

#endif
