#ifndef OnX_GuiPageViewer_h
#define OnX_GuiPageViewer_h

#include <string>

#include <Lib/smanip.h>
#include <Lib/Debug.h>

// Inventor :
#include <Inventor/errors/SoDebugError.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/events/SoEvent.h>
// HEPVis :
#include <HEPVis/SbPainterPS.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodes/SoViewportRegion.h>
#include <HEPVis/nodes/SoImageWriter.h>
#include <HEPVis/nodekits/SoPage.h>
#include <HEPVis/nodekits/SoRegion.h>
#include <HEPVis/actions/SoPainterAction.h>
#include <HEPVis/actions/SoGL2PSAction.h>

template <class SoGuiWidget, class SoGuiViewer,class SoGuiEvent> class GuiPageViewer;

template <class SoGuiWidget, class SoGuiViewer,class SoGuiEvent> 
class GuiPageSetRegionCallback : public virtual SbSetRegionCallback {
public:
  GuiPageSetRegionCallback(GuiPageViewer<SoGuiWidget,SoGuiViewer,SoGuiEvent>* aViewer):fViewer(aViewer){}
  virtual void setCamera(SoCamera* aCamera) {
    if(aCamera==fViewer->getCamera()) return; //done.
    fViewer->setCamera(aCamera);
  }
private:
  GuiPageViewer<SoGuiWidget,SoGuiViewer,SoGuiEvent>* fViewer;
};

template <class SoGuiWidget, class SoGuiViewer,class SoGuiEvent> 
class GuiPageViewer : public SoGuiViewer {
public:
  GuiPageViewer(SoGuiWidget aParent,const std::string& aName)
  :SoGuiViewer(aParent,aName.c_str(),TRUE),fPage(0) {
    Lib::Debug::increment("GuiPageViewer");
    // OSF1/cxx needs the "this->".
    const SbViewportRegion& vpRegion = this->getViewportRegion();
    SoGL2PSAction* action = new SoGL2PSAction(vpRegion);
    this->setGLRenderAction(action);

    fPage = new SoPage;
    fPage->ref();
    // SetRegionCallback object deleted by the SoPage.
    fPage->setSetRegionCallback
      (new GuiPageSetRegionCallback<SoGuiWidget,SoGuiViewer,SoGuiEvent>(this));
    
    this->setSceneGraph(fPage);
    
    fPage->highlightRegion(fPage->getRootRegion(0));

    // Disconnect headlight since all regions have their own ligh.
    // (Headlight will add a directinal light, then scene will
    // be more bright).
    // OSF1/cxx needs the "this->".
    this->setHeadlight(FALSE);
  }
  virtual ~GuiPageViewer() {
    Lib::Debug::decrement("GuiPageViewer");
    if(fPage) {
      fPage->setSetRegionCallback(0);
      fPage->unref();
    }
  }
  SoPage* getPage() const {return fPage;}
  bool write(const std::string& aWhat,
             const std::string& aFile,
             const std::string& aOptions) {
    //printf("debug : GuiPageViewerWrite : \"%s\" \"%s\"\n",
    //aWhat.c_str(),aFile.c_str());
    if(aWhat=="GIF") {
      writeGIF(aFile);
      return true;
    } else if(aWhat=="JPEG") {
      int quality;
      Lib::smanip::toint(aOptions,quality);
      writeJPEG(aFile,quality);
      return true;
    } else if(aWhat=="PS") {
      writePS(aFile);
      return true;
    } else if(aWhat=="GL2PS") {
      writeGL2(aFile,SoGL2PSAction::EPS,aOptions);
      return true;
    } else if(aWhat=="GL2TEX") {
      writeGL2(aFile,SoGL2PSAction::TEX,aOptions);
      return true;
    } else if(aWhat=="GL2PDF") {
      writeGL2(aFile,SoGL2PSAction::PDF,aOptions);
      return true;
    } else if(aWhat=="GL2SVG") {
      writeGL2(aFile,SoGL2PSAction::SVG,aOptions);
      return true;
    } else if(aWhat=="GL2PGF") {
      writeGL2(aFile,SoGL2PSAction::PGF,aOptions);
      return true;
    } else if(aWhat=="PSVec") {
      writePSVec(aFile,
                 this->getViewportRegion(),
                 this->getBackgroundColor(),
                 this->getSceneGraph());
      return true;
    }
    return false;
  }
  bool setParameter(const std::string& aWhat,
                    const std::string& aValue) {
    if( (aWhat=="view") && (aValue=="all")) {
      this->viewAll();
      return true;
    } else if( (aWhat=="scheduleRedraw") && (aValue=="")) {
      this->scheduleRedraw();
      return true;
    } else if( (aWhat=="render") && (aValue=="")) {
      this->render();
      return true;
    } else if( (aWhat=="animating") && (aValue=="stop")) {
      this->stopAnimating();
      return true;
    } else if(aWhat=="autoClipping") {
      bool value = true;
      if(!Lib::smanip::tobool(aValue,value)) value = true;
      if(value) {
        if(this->isAutoClipping()==FALSE) 
          this->setAutoClipping(TRUE);
      } else {
        if(this->isAutoClipping()==TRUE) 
          this->setAutoClipping(FALSE);
      }
      return true;
    } else if(aWhat=="decoration") {
      bool value = true;
      if(!Lib::smanip::tobool(aValue,value)) value = true;
      if(value) {
        if(this->isDecoration()==FALSE) 
          this->setDecoration(TRUE);
      } else {
        if(this->isDecoration()==TRUE) 
          this->setDecoration(FALSE);
      }
      return true;
    } else if(aWhat=="feedbackVisibility") {
      bool value = true;
      if(!Lib::smanip::tobool(aValue,value)) value = true;
      if(value) {
        if(this->isFeedbackVisible()==FALSE) 
          this->setFeedbackVisibility(TRUE);
      } else {
        if(this->isFeedbackVisible()==TRUE) 
          this->setFeedbackVisibility(FALSE);
      }
      return true;

    } else if(aWhat=="animationEnabled") {
      bool value = true;
      if(!Lib::smanip::tobool(aValue,value)) value = true;
      if(value) {
        if(this->isAnimationEnabled()==FALSE) 
          this->setAnimationEnabled(TRUE);
      } else {
        if(this->isAnimationEnabled()==TRUE) 
          this->setAnimationEnabled(FALSE);
      }
      return true;
    } else if(aWhat=="headLight") {
      bool value = true;
      if(!Lib::smanip::tobool(aValue,value)) value = true;
      if(value) {
        if(this->isHeadlight()==FALSE) 
          this->setHeadlight(TRUE); // Add a directional light !
      } else {
        if(this->isHeadlight()==TRUE) 
          this->setHeadlight(FALSE);
      }
      return true;
    } else if(aWhat=="smoothing") {
      SoGLRenderAction* action = this->getGLRenderAction();  
      if(!action) return false;
      bool value = true;
      if(!Lib::smanip::tobool(aValue,value)) value = true;
      if(value) {
        if(action->isSmoothing()==FALSE) action->setSmoothing(TRUE);
      } else {
        if(action->isSmoothing()==TRUE) action->setSmoothing(FALSE);
      }
      return true;
    } else if(aWhat=="viewing") {
      bool value = true;
      if(Lib::smanip::tobool(aValue,value)) {
        if(value) {
          if(this->isViewing()==FALSE) this->setViewing(TRUE);
        } else {
          if(this->isViewing()==TRUE) this->setViewing(FALSE);
        }
        return true;
      }
    } else if(aWhat=="backgroundColor") {
      double r,g,b;
      if(!Lib::smanip::torgb(aValue,r,g,b)) return false;
      this->setBackgroundColor(SbColor((float)r,(float)g,(float)b));
      return true;
    } else if(aWhat=="size") {
      std::vector<int> is;
      if(!Lib::smanip::toints(aValue,is)) return false;
      if(is.size()!=2) return false;
      this->setSize(SbVec2s((short)is[0],(short)is[1]));
      return true;
    } else if(aWhat=="title") {
      this->setTitle(aValue.c_str());
      return true;
    } else {
      SoDebugError::postInfo("GuiPageViewer::setParameter",
                             "\"%s\" not a parameter.",aWhat.c_str());
    }
    return false;
  }
  void writeGIF(const std::string& aFile) {
    if(!fPage) return;
    SoRegion* soRegion = fPage->getHighlightedRegion();
    if(!soRegion) return;

    SoImageWriter* imageWriter = fPage->getImageWriter();
    imageWriter->fileName.setValue(aFile.c_str());
    imageWriter->format.setValue(SoImageWriter::GIF);
    imageWriter->enable();

    SoViewportRegion* viewportRegion = soRegion->getViewportRegion();
    viewportRegion->highlighted.setValue(FALSE);
    this->render();
    viewportRegion->highlighted.setValue(TRUE);

    imageWriter->disable();
    if(imageWriter->getStatus()) {
      SoDebugError::postInfo("","%s produced.",
                             imageWriter->fileName.getValue().getString());
    } else {
      SoDebugError::postInfo("","%s not produced.",
                             imageWriter->fileName.getValue().getString());
    }
  }
  void writeJPEG(const std::string& aFile,int aQuality = 60) {
    if(!fPage) return;
    SoRegion* soRegion = fPage->getHighlightedRegion();
    if(!soRegion) return;
    SoImageWriter* imageWriter = fPage->getImageWriter();
    imageWriter->fileName.setValue(aFile.c_str());
    imageWriter->format.setValue(SoImageWriter::JPEG);
    imageWriter->quality.setValue(aQuality);
    imageWriter->enable();

    SoViewportRegion* viewportRegion = soRegion->getViewportRegion();
    viewportRegion->highlighted.setValue(FALSE);
    this->render();
    viewportRegion->highlighted.setValue(TRUE);

    imageWriter->disable();
    if(imageWriter->getStatus()) {
      SoDebugError::postInfo("","%s produced.",
                             imageWriter->fileName.getValue().getString());
    } else {
      SoDebugError::postInfo("","%s not produced.",
                             imageWriter->fileName.getValue().getString());
    }
  }
  void writePS(const std::string& aFile) {
    if(!fPage) return;
    SoRegion* soRegion = fPage->getHighlightedRegion();
    if(!soRegion) return;
    SoImageWriter* imageWriter = fPage->getImageWriter();
    imageWriter->fileName.setValue(aFile.c_str());
    imageWriter->format.setValue(SoImageWriter::POST_SCRIPT);
    imageWriter->enable();

    SoViewportRegion* viewportRegion = soRegion->getViewportRegion();
    viewportRegion->highlighted.setValue(FALSE);
    this->render();
    viewportRegion->highlighted.setValue(TRUE);

    imageWriter->disable();
    if(imageWriter->getStatus()) {
      SoDebugError::postInfo("","%s produced.",
                             imageWriter->fileName.getValue().getString());
    } else {
      SoDebugError::postInfo("","%s not produced.",
                             imageWriter->fileName.getValue().getString());
    }
  }  
  void writePSVec(const std::string& aFile
                  ,const SbViewportRegion& aRegion
                  ,const SbColor& aBackground
                  ,SoNode* aSceneGraph) {
    // Open file :
    SbPainterPS painterPS;
    painterPS.openFileForWriting(aFile.c_str());
    // Produce view :
    SoPainterAction* action = new SoPainterAction(aRegion,&painterPS);
    action->setBackgroundColor(aBackground);
    action->apply(aSceneGraph);
    SoDebugError::postInfo("","View put in %s.",painterPS.getFileName());
    delete action;
    // Close file :
    if(painterPS.getStream()) 
      SoDebugError::postInfo("","%s produced.",painterPS.getFileName());
    else 
      SoDebugError::postInfo("","%s not produced.",painterPS.getFileName());
    painterPS.closeStream();
  }
  void writeGL2(const std::string& aFile,
                SoGL2PSAction::Format aFormat,
                const std::string& aOptions) {
    SoGLRenderAction* action = this->getGLRenderAction();  
    if(action->isOfType(SoGL2PSAction::getClassTypeId())) { 
      SoGL2PSAction* gl2ps_action = (SoGL2PSAction*)action;
      gl2ps_action->setPageOptions(aOptions.c_str());
      gl2ps_action->setFileFormat(aFormat);
      gl2ps_action->setFileName(aFile.c_str());
      gl2ps_action->enableFileWriting();
      this->render();
      gl2ps_action->disableFileWriting();
    }
  }
  const SoEvent* soEvent(SoGuiEvent* aEvent) {return getSoEvent(aEvent);}
  void collect(int aX,int aY,int aW,int aH){
    // OSF1/cxx needs the "this->".
    const SbViewportRegion& vpRegion = this->getViewportRegion();
    fPage->collect(aX,aY,aW,aH,vpRegion);
  }
  void map(const SbVec2s& aPos,  // Rectangle position (GL coordinates).
           const SbVec2s& aSize, // Rectangle size.
           const SbVec2s& aWindow) {
    SoRegion* soRegion = fPage->getHighlightedRegion();
    if(!soRegion) return;
    soRegion->pushCamera();
    soRegion->viewRectangle(aPos,aSize,aWindow);
  }
  bool isInCurrentRegion(const SbVec2s& aPos) {
    // aPos is in GL coordinates.
    SoRegion* soRegion = fPage->getHighlightedRegion();
    if(!soRegion) return false;
    SoViewportRegion* soViewportRegion = soRegion->getViewportRegion();
    int rx,ry;
    soViewportRegion->getPositionPixels(rx,ry);
    unsigned int rw,rh;
    soViewportRegion->getSizePixels(rw,rh);
    if(aPos[0]<int(rx)) return false;
    if(aPos[0]>=int(rx+rw)) return false;
    if(aPos[1]<int(ry)) return false;
    if(aPos[1]>=int(ry+rh)) return false;
    return true;
  }
  void pushCamera() {
    SoRegion* soRegion = fPage->getHighlightedRegion();
    if(!soRegion) return;
    soRegion->pushCamera();
  }
  void popCamera() {
    SoRegion* soRegion = fPage->getHighlightedRegion();
    if(!soRegion) return;
    soRegion->popCamera();
  }
private:
  SoPage* fPage;
};

#endif
