#ifndef OnX_GuiViewer_h
#define OnX_GuiViewer_h

#include <Lib/smanip.h>

// Inventor :
#include <Inventor/errors/SoDebugError.h>

// HEPVis :
#include <HEPVis/SbPainterPS.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodes/SoViewportRegion.h>
#include <HEPVis/nodes/SoImageWriter.h>
#include <HEPVis/nodekits/SoPage.h>
#include <HEPVis/nodekits/SoRegion.h>
#include <HEPVis/actions/SoPainterAction.h>
#include <HEPVis/actions/SoGL2PSAction.h>

template <class SoGuiViewer> 
class GuiPageSetRegionCallback : public virtual SbSetRegionCallback {
public:
  GuiPageSetRegionCallback(SoGuiViewer& aViewer):fViewer(aViewer){}
  virtual void setCamera(SoCamera* aCamera) {
    if(aCamera==fViewer.getCamera()) return; //done.
    fViewer.setCamera(aCamera);
  }
private:
  SoGuiViewer& fViewer;
};

template <class SoGuiViewer> 
void GuiViewer_initialize(SoGuiViewer& aThis) {
  const SbViewportRegion& vpRegion = aThis.getViewportRegion();
  SoGL2PSAction* action = new SoGL2PSAction(vpRegion);
  aThis.setGLRenderAction(action);

  SoPage* soPage = new SoPage;
  soPage->ref();
  // SetRegionCallback object deleted by the SoPage.
  soPage->setSetRegionCallback
    (new GuiPageSetRegionCallback<SoGuiViewer>(aThis));
    
  aThis.setSceneGraph(soPage);
    
  soPage->highlightRegion(soPage->getRootRegion(0));

  // Disconnect headlight since all regions have their own light.
  // (Headlight will add a directinal light, then scene will
  // be more bright).
  // OSF1/cxx needs the "this->".
  aThis.setHeadlight(FALSE);
}

template <class SoGuiViewer> 
SoPage* GuiViewer_SoPage(SoGuiViewer& aThis) {
  SoNode* node = aThis.getSceneGraph();
  if(!node) return 0;        
  if(node->isOfType(SoPage::getClassTypeId())==FALSE) return 0;
  return (SoPage*)node;
}

template <class SoGuiViewer> 
bool GuiViewer_finalize(SoGuiViewer& aThis) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  soPage->setSetRegionCallback(0);
  soPage->unref();
  return true;
}

template <class SoGuiViewer> 
bool GuiViewer_writeGIF(SoGuiViewer& aThis,const std::string& aFile) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;

  SoImageWriter* imageWriter = soPage->getImageWriter();
  imageWriter->fileName.setValue(aFile.c_str());
  imageWriter->format.setValue(SoImageWriter::GIF);
  imageWriter->enable();

  SoViewportRegion* viewportRegion = soRegion->getViewportRegion();
  viewportRegion->highlighted.setValue(FALSE);
  aThis.render();
  viewportRegion->highlighted.setValue(TRUE);

  imageWriter->disable();
  if(imageWriter->getStatus()) {
    SoDebugError::postInfo("","%s produced.",
                           imageWriter->fileName.getValue().getString());
    return true;
  } else {
    SoDebugError::postInfo("","%s not produced.",
                           imageWriter->fileName.getValue().getString());
    return false;
  }
}

template <class SoGuiViewer> 
bool GuiViewer_writeJPEG(SoGuiViewer& aThis,const std::string& aFile,int aQuality = 60) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  SoImageWriter* imageWriter = soPage->getImageWriter();
  imageWriter->fileName.setValue(aFile.c_str());
  imageWriter->format.setValue(SoImageWriter::JPEG);
  imageWriter->quality.setValue(aQuality);
  imageWriter->enable();

  SoViewportRegion* viewportRegion = soRegion->getViewportRegion();
  viewportRegion->highlighted.setValue(FALSE);
  aThis.render();
  viewportRegion->highlighted.setValue(TRUE);

  imageWriter->disable();
  if(imageWriter->getStatus()) {
    SoDebugError::postInfo("","%s produced.",
                           imageWriter->fileName.getValue().getString());
    return true;
  } else {
    SoDebugError::postInfo("","%s not produced.",
                           imageWriter->fileName.getValue().getString());
    return false;
  }
}

template <class SoGuiViewer> 
bool GuiViewer_writePS(SoGuiViewer& aThis,const std::string& aFile) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  SoImageWriter* imageWriter = soPage->getImageWriter();
  imageWriter->fileName.setValue(aFile.c_str());
  imageWriter->format.setValue(SoImageWriter::POST_SCRIPT);
  imageWriter->enable();

  SoViewportRegion* viewportRegion = soRegion->getViewportRegion();
  viewportRegion->highlighted.setValue(FALSE);
  aThis.render();
  viewportRegion->highlighted.setValue(TRUE);

  imageWriter->disable();
  if(imageWriter->getStatus()) {
    SoDebugError::postInfo("","%s produced.",
                           imageWriter->fileName.getValue().getString());
    return true;
  } else {
    SoDebugError::postInfo("","%s not produced.",
                           imageWriter->fileName.getValue().getString());
    return false;
  }
}  

bool GuiViewer_writePSVec(
 const std::string& aFile
,const SbViewportRegion& aRegion
,const SbColor& aBackground
,SoNode* aSceneGraph) {
  // Open file :
  SbPainterPS painterPS;
  painterPS.openFileForWriting(aFile.c_str());
  // Produce view :
  SoPainterAction* action = new SoPainterAction(aRegion,&painterPS);
  action->setBackgroundColor(aBackground);
  action->apply(aSceneGraph);
  SoDebugError::postInfo("","View put in %s.",painterPS.getFileName());
  delete action;
  // Close file :
  bool status;
  if(painterPS.getStream()) {
    SoDebugError::postInfo("","%s produced.",painterPS.getFileName());
    status = true;
  } else {
    SoDebugError::postInfo("","%s not produced.",painterPS.getFileName());
    status = false;
  }
  painterPS.closeStream();
  return status;
}

template <class SoGuiViewer> 
bool GuiViewer_writeGL2(SoGuiViewer& aThis,
                        const std::string& aFile,
                        SoGL2PSAction::Format aFormat,
                        const std::string& aOptions) {
  SoGLRenderAction* action = aThis.getGLRenderAction();  
  if(action->isOfType(SoGL2PSAction::getClassTypeId())) { 
    SoGL2PSAction* gl2ps_action = (SoGL2PSAction*)action;
    gl2ps_action->setPageOptions(aOptions.c_str());
    gl2ps_action->setFileFormat(aFormat);
    gl2ps_action->setFileName(aFile.c_str());
    gl2ps_action->enableFileWriting();
    aThis.render();
    gl2ps_action->disableFileWriting();
    return true;
  } else {
    return false;
  }
}

template <class SoGuiViewer> 
bool GuiViewer_writeScene(SoGuiViewer& aThis,
                          const std::string& aFile,
                          const std::string& aFormat,
                          const std::string& aWhat,
                          bool aBinary,
                          bool aAltRep) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  SoNode* node = soRegion->getWriteNode(aWhat.c_str());
  if(!node) return false;
  if(SoTools::write(*node,
                    aFile.c_str(),
                    aFormat.c_str(),
                    aBinary?TRUE:FALSE,
                    aAltRep?TRUE:FALSE)==FALSE) {
    SoDebugError::postInfo("","Can't write file \"%s\".",aFile.c_str());
    return false;
  }
  return true;
}

template <class SoGuiViewer> 
bool GuiViewer_write(SoGuiViewer& aThis,
                     const std::string& aWhat,
                     const std::string& aFile,
                     const std::string& aOptions) {
    //printf("debug : GuiViewer_write : \"%s\" \"%s\"\n",
    //aWhat.c_str(),aFile.c_str());
    if(aWhat=="GIF") {
      return GuiViewer_writeGIF(aThis,aFile);
    } else if(aWhat=="JPEG") {
      int quality;
      Lib::smanip::toint(aOptions,quality);
      return GuiViewer_writeJPEG(aThis,aFile,quality);
    } else if(aWhat=="PS") {
      GuiViewer_writePS(aThis,aFile);
      return true;
    } else if(aWhat=="GL2PS") {
      return GuiViewer_writeGL2(aThis,aFile,SoGL2PSAction::EPS,aOptions);
    } else if(aWhat=="GL2TEX") {
      return GuiViewer_writeGL2(aThis,aFile,SoGL2PSAction::TEX,aOptions);
    } else if(aWhat=="GL2PDF") {
      return GuiViewer_writeGL2(aThis,aFile,SoGL2PSAction::PDF,aOptions);
    } else if(aWhat=="GL2SVG") {
      return GuiViewer_writeGL2(aThis,aFile,SoGL2PSAction::SVG,aOptions);
    } else if(aWhat=="GL2PGF") {
      return GuiViewer_writeGL2(aThis,aFile,SoGL2PSAction::PGF,aOptions);
    } else if(aWhat=="PSVec") {
      return GuiViewer_writePSVec(aFile,
                 aThis.getViewportRegion(),
                 aThis.getBackgroundColor(),
                 aThis.getSceneGraph());
    } else if( (aWhat=="HIV") || (aWhat=="IV") || (aWhat=="INVENTOR")  ||
               (aWhat=="WRL") || (aWhat=="VRML2") || 
               (aWhat=="POV") ){

      std::string format(aWhat);
      Lib::smanip::tolowercase(format);

      bool binary = false;
      bool altRep = true;
      std::string what = "scene"; //scene staticScene dynamicScene region page

     {std::vector< std::pair<std::string,std::string> > opts;
      if(!Lib::smanip::options(aOptions,opts)) return false;
      std::vector< std::pair<std::string,std::string> >::const_iterator it;
      for(it=opts.begin();it!=opts.end();++it) {
        if((*it).first=="binary") {
          if(!Lib::smanip::tobool((*it).second,binary)) return false;
        } else if((*it).first=="altRep") {
          if(!Lib::smanip::tobool((*it).second,altRep)) return false;
        } else if((*it).first=="what") {
          what = (*it).second;
        }
      }}

      return GuiViewer_writeScene(aThis,aFile,format,what,binary,altRep);
    }
    return false;
}

template <class SoGuiViewer> 
bool GuiViewer_collect(SoGuiViewer& aThis,int aX,int aY,int aW,int aH){
  const SbViewportRegion& vpRegion = aThis.getViewportRegion();
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  soPage->collect(aX,aY,aW,aH,vpRegion);
  return true;
}

template <class SoGuiViewer> 
bool GuiViewer_map(SoGuiViewer& aThis,
                 const SbVec2s& aPos,  // Rectangle position (GL coordinates).
                 const SbVec2s& aSize, // Rectangle size.
                 const SbVec2s& aWindow) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  soRegion->pushCamera();
  soRegion->viewRectangle(aPos,aSize,aWindow);
  return true;
}

template <class SoGuiViewer> 
bool GuiViewer_isInCurrentRegion(SoGuiViewer& aThis,const SbVec2s& aPos) {
  // aPos is in GL coordinates.
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  SoViewportRegion* soViewportRegion = soRegion->getViewportRegion();
  int rx,ry;
  soViewportRegion->getPositionPixels(rx,ry);
  unsigned int rw,rh;
  soViewportRegion->getSizePixels(rw,rh);
  if(aPos[0]<int(rx)) return false;
  if(aPos[0]>=int(rx+rw)) return false;
  if(aPos[1]<int(ry)) return false;
  if(aPos[1]>=int(ry+rh)) return false;
  return true;
}

template <class SoGuiViewer> 
bool GuiViewer_pushCamera(SoGuiViewer& aThis) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  soRegion->pushCamera();
  return true;
}

template <class SoGuiViewer> 
bool GuiViewer_popCamera(SoGuiViewer& aThis) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  soRegion->popCamera();
  return true;
}

// FullViewer set parameter :
template <class SoGuiViewer> 
bool GuiViewer_setParameter(SoGuiViewer& aThis,
                            const std::string& aWhat,
                            const std::string& aValue) {
  if( (aWhat=="view") && (aValue=="all")) {
    aThis.viewAll();
    return true;
  } else if( (aWhat=="scheduleRedraw") && (aValue=="")) {
    aThis.scheduleRedraw();
    return true;
  } else if( (aWhat=="render") && (aValue=="")) {
    aThis.render();
    return true;
  } else if(aWhat=="autoClipping") {
    bool value = true;
    if(!Lib::smanip::tobool(aValue,value)) value = true;
    if(value) {
      if(aThis.isAutoClipping()==FALSE) 
        aThis.setAutoClipping(TRUE);
    } else {
      if(aThis.isAutoClipping()==TRUE) 
        aThis.setAutoClipping(FALSE);
    }
    return true;
  } else if(aWhat=="decoration") {
    bool value = true;
    if(!Lib::smanip::tobool(aValue,value)) value = true;
    if(value) {
      if(aThis.isDecoration()==FALSE) 
        aThis.setDecoration(TRUE);
    } else {
      if(aThis.isDecoration()==TRUE) 
        aThis.setDecoration(FALSE);
    }
    return true;
  } else if(aWhat=="headLight") {
    bool value = true;
    if(!Lib::smanip::tobool(aValue,value)) value = true;
    if(value) {
      if(aThis.isHeadlight()==FALSE) 
        aThis.setHeadlight(TRUE); // Add a directional light !
    } else {
      if(aThis.isHeadlight()==TRUE) 
        aThis.setHeadlight(FALSE);
    }
    return true;
  } else if(aWhat=="smoothing") {
    SoGLRenderAction* action = aThis.getGLRenderAction();  
    if(!action) return false;
    bool value = true;
    if(!Lib::smanip::tobool(aValue,value)) value = true;
    if(value) {
      if(action->isSmoothing()==FALSE) action->setSmoothing(TRUE);
    } else {
      if(action->isSmoothing()==TRUE) action->setSmoothing(FALSE);
    }
    return true;
  } else if(aWhat=="viewing") {
    bool value = true;
    if(Lib::smanip::tobool(aValue,value)) {
      if(value) {
        if(aThis.isViewing()==FALSE) aThis.setViewing(TRUE);
      } else {
        if(aThis.isViewing()==TRUE) aThis.setViewing(FALSE);
      }
      return true;
    }
  } else if(aWhat=="backgroundColor") {
    double r,g,b;
    if(!Lib::smanip::torgb(aValue,r,g,b)) return false;
    aThis.setBackgroundColor(SbColor((float)r,(float)g,(float)b));
    return true;
  } else if(aWhat=="size") {
    std::vector<int> is;
    if(!Lib::smanip::toints(aValue,is)) return false;
    if(is.size()!=2) return false;
    aThis.setSize(SbVec2s((short)is[0],(short)is[1]));
    return true;
  } else if(aWhat=="title") {
    aThis.setTitle(aValue.c_str());
    return true;
  }
  return false;
}

// ExaminerViewer set parameter :
template <class SoGuiViewer> 
bool GuiExaminerViewer_setParameter(SoGuiViewer& aThis,
                                    const std::string& aWhat,
                                    const std::string& aValue) {
  if( (aWhat=="animating") && (aValue=="stop")) {
    aThis.stopAnimating();
    return true;
  } else if(aWhat=="feedbackVisibility") {
    bool value = true;
    if(!Lib::smanip::tobool(aValue,value)) value = true;
    if(value) {
      if(aThis.isFeedbackVisible()==FALSE) 
        aThis.setFeedbackVisibility(TRUE);
    } else {
      if(aThis.isFeedbackVisible()==TRUE) 
        aThis.setFeedbackVisibility(FALSE);
    }
    return true;
  } else if(aWhat=="animationEnabled") {
    bool value = true;
    if(!Lib::smanip::tobool(aValue,value)) value = true;
    if(value) {
      if(aThis.isAnimationEnabled()==FALSE) 
        aThis.setAnimationEnabled(TRUE);
    } else {
      if(aThis.isAnimationEnabled()==TRUE) 
        aThis.setAnimationEnabled(FALSE);
    }
    return true;
  } else if(GuiViewer_setParameter(aThis,aWhat,aValue)) {
    return true;
  }
  return false;
}


/*
  const SoEvent* soEvent(SoGuiEvent* aEvent) {return getSoEvent(aEvent);}
*/

#endif
